from .bootstrap import _BootstrapMainModule
from .parser import parse
from .query import Query, QueryAsync, QueryBuilder
from .utils import get_async_client, get_session, recursive_merge
from .version import version

# Private methods


_bootstrap = _BootstrapMainModule()
_query_builder = QueryBuilder(_bootstrap)


# Expose class methods as the public API


bootstrap = _bootstrap.bootstrap
bootstrap_async = _bootstrap.bootstrap_async
is_bootstrapped = _bootstrap.is_bootstrapped
clear_bootstrapping = _bootstrap.clear_bootstrapping
save_bootstrap_data = _bootstrap.save_bootstrap_data
load_bootstrap_data = _bootstrap.load_bootstrap_data
bootstrap_is_older_than = _bootstrap.bootstrap_is_older_than
build_query = _query_builder.build


# Query helpers


def _asn(as_number, rir=None, raw=False, session=None, async_client=None):
    is_async = async_client is not None
    method, url, _ = build_query(query_type='asn', query_value=as_number, rir=rir)

    if is_async:
        q = QueryAsync(async_client, method, url)
    else:
        q = Query(session, method, url)
    response = yield q
    yield response if raw else parse(_bootstrap, 'autnum', as_number, response)


def asn(as_number, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl)
    gen = _asn(as_number, rir, raw, session=session)
    q: Query = next(gen)
    resp: dict = gen.send(q.request())
    gen.close()
    return resp


async def asn_async(as_number, rir=None, raw=False, allow_insecure_ssl=False, async_client=None):
    async_client = get_async_client(async_client, allow_insecure_ssl)
    gen = _asn(as_number, rir, raw, async_client=async_client)
    q: QueryAsync = next(gen)
    resp: dict = gen.send(await q.request())
    gen.close()
    return resp


def _domain(domain_name, raw=False, session=None, follow_related=True, async_client=None, is_async=False):
    is_async = async_client is not None
    method, url, _ = build_query(query_type='domain', query_value=domain_name)

    if is_async:
        q = QueryAsync(async_client, method, url)
    else:
        q = Query(session, method, url)

    response = yield q

    if raw:
        yield response
    if follow_related:
        # Attempt to follow the 'related' or 'registration' links if the TLD has
        # an upstream RDAP endpoint that may have more information
        relresponse = None
        for link in response.get('links', []):
            rel = link.get('rel', '')
            if rel in ('related', 'registration'):
                relhref = link.get('href', '')
                if relhref:
                    if is_async:
                        relq = QueryAsync(async_client, method, relhref)
                    else:
                        relq= Query(session, method, relhref)
                    yield
                    relresponse = yield relq
                    break
        if relresponse:
            # Overlay the related response over the original response
            recursive_merge(response, relresponse)
    yield parse(_bootstrap, 'domain', domain_name, response)


def domain(domain_name, raw=False, allow_insecure_ssl=False, session=None, follow_related=True):
    session = get_session(session, allow_insecure_ssl)
    gen = _domain(domain_name, raw, session, follow_related, None)
    resp: dict = None

    q: Query
    for q in gen:
        req_resp: dict = q.request()
        resp = gen.send(req_resp)

        if resp is req_resp:
            gen.close()
            break

    return resp


async def domain_async(domain_name, raw=False, allow_insecure_ssl=False, async_client=None, follow_related=True):
    async_client = get_async_client(async_client, allow_insecure_ssl)
    gen = _domain(domain_name, raw, None, follow_related, async_client)
    resp: dict = None

    q: QueryAsync
    for q in gen:
        req_resp: dict = await q.request()
        resp = gen.send(req_resp)

        # happens when raw=true
        if resp is req_resp:
            gen.close()
            break

    return resp


def _ip(ip_address_or_network, rir=None, raw=False, session=None, async_client=None, is_async=False):
    is_async = async_client is not None
    method, url, _ = build_query(query_type='ip', query_value=ip_address_or_network, rir=rir)

    if is_async:
        q = QueryAsync(async_client, method, url)
    else:
        q = Query(session, method, url)
    response = yield q
    yield response if raw else parse(_bootstrap, 'ip', ip_address_or_network, response)


def ip(ip_address_or_network, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl)
    gen = _ip(ip_address_or_network, rir, raw, session, None)
    q: Query = next(gen)
    resp: dict = gen.send(q.request())
    gen.close()
    return resp


async def ip_async(ip_address_or_network, rir=None, raw=False, allow_insecure_ssl=False, async_client=None):
    async_client = get_async_client(async_client, allow_insecure_ssl)
    gen = _ip(ip_address_or_network, rir, raw, None, async_client)
    q: QueryAsync = next(gen)
    resp: dict = gen.send(await q.request())
    gen.close()
    return resp


def _entity(entity_handle, rir=None, raw=False, session=None, async_client=None, is_async=False):
    is_async = async_client is not None
    method, url, _ = build_query(query_type='entity', query_value=entity_handle, rir=rir)

    if is_async:
        q = QueryAsync(async_client, method, url)
    else:
        q = Query(session, method, url)
    response = yield q
    yield response if raw else parse(_bootstrap, 'entity', entity_handle, response)


def entity(entity_handle, rir=None, raw=False, allow_insecure_ssl=False, session=None):
    session = get_session(session, allow_insecure_ssl)
    gen = _entity(entity_handle, rir, raw, session, None)
    q: Query = next(gen)
    resp: dict = gen.send(q.request())
    gen.close()
    return resp


async def entity_async(ip_address_or_network, rir=None, raw=False, allow_insecure_ssl=False, async_client=None):
    async_client = get_async_client(async_client, allow_insecure_ssl)
    gen = _entity(ip_address_or_network, rir, raw, None, async_client)
    q: QueryAsync = next(gen)
    resp: dict = gen.send(await q.request())
    gen.close()
    return resp



import json
import unittest
from datetime import datetime, timezone
from ipaddress import IPv4Network, IPv6Network
from pathlib import Path

import pytest
import responses
import yaml
from dateutil.tz import tzoffset
from responses import _recorder

import whoisit

BASE_DIR = Path(__file__).resolve().parent
RESPONSES = BASE_DIR / "responses"
UTC = timezone.utc


class SyncPublicInterfaceTestCase(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()

    def test_version(self):
        self.assertRegex(whoisit.version, r"\d+\.\d+\.\d+")

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_bootstrap_interface(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'asn-1.yaml')
    def test_asn_interface_1(self):
        responses._add_from_file(RESPONSES / "asn-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.asn(13335)
        self.assertEqual(resp["type"], "autnum")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/autnum/13335")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["asn_range"], [13335, 13335])
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(resp["copyright_notice"], "Copyright 1997-2024, American Registry for Internet Numbers, Ltd.")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["registration_date"], datetime(2010, 7, 14, 18, 35, 57, tzinfo=tzoffset(None, -14400)))
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["abuse"], list))
        self.assertTrue(len(resp["entities"]["abuse"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["noc"], list))
        self.assertTrue(len(resp["entities"]["noc"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'domain-1.yaml')
    def test_domain_interface_1(self):
        responses._add_from_file(RESPONSES / "domain-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.domain("google.com")
        self.assertEqual(resp["type"], "domain")
        self.assertEqual(resp["name"], "google.com")
        self.assertEqual(resp["handle"], "2138514_DOMAIN_COM-VRSN")
        self.assertEqual(resp["rir"], "")
        self.assertEqual(resp["registration_date"], datetime(1997, 9, 15, 7, 0, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.markmonitor.com/rdap/domain/google.com")
        self.assertEqual(
            resp["terms_of_service_url"],
            "https://www.markmonitor.com/legal/domain-management-terms-and-conditions",
        )
        self.assertEqual(resp["whois_server"], "whois.markmonitor.com")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["dnssec"], False)
        self.assertEqual(resp["nameservers"], ["ns1.google.com", "ns2.google.com", "ns3.google.com", "ns4.google.com"])
        self.assertEqual(
            resp["status"],
            [
                "client update prohibited",
                "client transfer prohibited",
                "client delete prohibited",
                "server update prohibited",
                "server transfer prohibited",
                "server delete prohibited",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["registrar"], list))
        self.assertTrue(len(resp["entities"]["registrar"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertEqual(
            resp["entities"]["registrant"][0]["address"],
            {
                "country": "US",
                "ext_address": "",
                "locality": "",
                "po_box": "",
                "postal_code": "",
                "region": "CA",
                "street_address": "",
            },
        )

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v4-1.yaml')
    def test_ip_interface_v4_1(self):
        responses._add_from_file(RESPONSES / "ip-v4-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("1.1.1.1")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v6-1.yaml')
    def test_ip_interface_v6_1(self):
        responses._add_from_file(RESPONSES / "ip-v6-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("2606:4700:4700::1111")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v4-cidr-1.yaml')
    def test_ip_interface_v4_cidr_1(self):
        responses._add_from_file(RESPONSES / "ip-v4-cidr-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("1.1.1.0/24")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["assignment_type"], "assigned portable")
        self.assertEqual(resp["country"], "AU")
        self.assertEqual(resp["ip_version"], 4)
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'ip-v6-cidr-1.yaml')
    def test_ip_interface_v6_cidr_1(self):
        responses._add_from_file(RESPONSES / "ip-v6-cidr-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.ip("2606:4700::/32")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'entity-1.yaml')
    def test_entity_interface_1(self):
        responses._add_from_file(RESPONSES / "entity-1.yaml")
        whoisit.bootstrap()
        resp = whoisit.entity("GOVI", rir="arin")
        self.assertEqual(resp["type"], "entity")
        self.assertEqual(resp["name"], "Govital Internet Inc.")
        self.assertEqual(resp["handle"], "GOVI")
        self.assertEqual(resp["parent_handle"], "")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2001, 5, 8, 0, 0, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["expiration_date"], None)
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/entity/GOVI")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["http://www.govital.net\r", "Standard NOC hours are 10am to 6pm EST M-F"]
        )
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_save_bootstrap_data(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

        result = whoisit.save_bootstrap_data()
        self.assertTrue(isinstance(result, str))

        data = json.loads(result)
        self.assertTrue(bool(data))

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_save_bootstrap_data_no_json(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data(as_json=False)
        self.assertTrue(isinstance(data, dict))
        self.assertTrue(bool(data))

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_load_bootstrap_data(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data()
        self.assertTrue(isinstance(data, str))

        whoisit.clear_bootstrapping()

        whoisit.load_bootstrap_data(data)
        self.assertTrue(whoisit.is_bootstrapped())

    @responses.activate
    # @_recorder.record(file_path=RESPONSES / 'boostrap.yaml')
    def test_load_bootstrap_data_no_json(self):
        responses._add_from_file(RESPONSES / "boostrap.yaml")
        whoisit.bootstrap()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data(as_json=False)
        self.assertTrue(isinstance(data, dict))

        whoisit.clear_bootstrapping()

        whoisit.load_bootstrap_data(data, from_json=False)
        self.assertTrue(whoisit.is_bootstrapped())


#############################################
################### Async ###################
#############################################


def load_sync_responses_to_httpx_mock(file, httpx_mock):
    with open(file) as fh:
        data = yaml.safe_load(fh)

    for item in data["responses"]:
        response = item["response"]
        httpx_mock.add_response(
            url=response["url"],
            method=response["method"],
            status_code=response["status"],
            text=response["body"],
            headers=response["headers"],
        )


@pytest.fixture(scope="function")
def mock_httpx(request, httpx_mock):
    request.cls.httpx_mock = httpx_mock


class AsyncPublicInterfaceTestCase(unittest.IsolatedAsyncioTestCase):
    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()

    def test_version(self):
        self.assertRegex(whoisit.version, r"\d+\.\d+\.\d+")

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_bootstrap_interface(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_asn_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "asn-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.asn_async(13335)
        self.assertEqual(resp["type"], "autnum")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/autnum/13335")
        self.assertEqual(resp["handle"], "AS13335")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["asn_range"], [13335, 13335])
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(resp["copyright_notice"], "Copyright 1997-2024, American Registry for Internet Numbers, Ltd.")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["registration_date"], datetime(2010, 7, 14, 18, 35, 57, tzinfo=tzoffset(None, -14400)))
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["abuse"], list))
        self.assertTrue(len(resp["entities"]["abuse"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["noc"], list))
        self.assertTrue(len(resp["entities"]["noc"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_domain_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "domain-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.domain_async("google.com")
        self.assertEqual(resp["type"], "domain")
        self.assertEqual(resp["name"], "google.com")
        self.assertEqual(resp["handle"], "2138514_DOMAIN_COM-VRSN")
        self.assertEqual(resp["rir"], "")
        self.assertEqual(resp["registration_date"], datetime(1997, 9, 15, 7, 0, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.markmonitor.com/rdap/domain/google.com")
        self.assertEqual(
            resp["terms_of_service_url"],
            "https://www.markmonitor.com/legal/domain-management-terms-and-conditions",
        )
        self.assertEqual(resp["whois_server"], "whois.markmonitor.com")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["dnssec"], False)
        self.assertEqual(resp["nameservers"], ["ns1.google.com", "ns2.google.com", "ns3.google.com", "ns4.google.com"])
        self.assertEqual(
            resp["status"],
            [
                "client update prohibited",
                "client transfer prohibited",
                "client delete prohibited",
                "server update prohibited",
                "server transfer prohibited",
                "server delete prohibited",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["registrar"], list))
        self.assertTrue(len(resp["entities"]["registrar"]) > 0)
        self.assertTrue(isinstance(resp["entities"]["registrant"], list))
        self.assertTrue(len(resp["entities"]["registrant"]) > 0)
        self.assertEqual(
            resp["entities"]["registrant"][0]["address"],
            {
                "country": "US",
                "ext_address": "",
                "locality": "",
                "po_box": "",
                "postal_code": "",
                "region": "CA",
                "street_address": "",
            },
        )

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v4_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v4-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("1.1.1.1")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v6_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v6-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("2606:4700:4700::1111")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v4_cidr_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v4-cidr-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("1.1.1.0/24")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "APNIC-LABS")
        self.assertEqual(resp["handle"], "1.1.1.0 - 1.1.1.255")
        self.assertEqual(resp["rir"], "apnic")
        self.assertEqual(resp["registration_date"], datetime(2011, 8, 10, 23, 12, 35, tzinfo=UTC))
        self.assertEqual(resp["url"], "https://rdap.apnic.net/ip/1.1.1.0/24")
        self.assertEqual(resp["terms_of_service_url"], "http://www.apnic.net/db/dbcopyright.html")
        self.assertEqual(resp["whois_server"], "whois.apnic.net")
        self.assertEqual(resp["copyright_notice"], "")
        self.assertEqual(resp["assignment_type"], "assigned portable")
        self.assertEqual(resp["country"], "AU")
        self.assertEqual(resp["ip_version"], 4)
        self.assertEqual(resp["network"], IPv4Network("1.1.1.0/24"))
        self.assertEqual(
            resp["description"],
            [
                "APNIC and Cloudflare DNS Resolver project",
                "Routed globally by AS13335/Cloudflare",
                "Research prefix for APNIC Labs",
            ],
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_ip_interface_v6_cidr_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "ip-v6-cidr-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.ip_async("2606:4700::/32")
        self.assertEqual(resp["type"], "ip network")
        self.assertEqual(resp["name"], "CLOUDFLARENET")
        self.assertEqual(resp["handle"], "NET6-2606-4700-1")
        self.assertEqual(resp["network"], IPv6Network("2606:4700::/32"))
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2011, 11, 1, 15, 59, 58, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/ip/2606:4700::")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["All Cloudflare abuse reporting can be done via https://www.cloudflare.com/abuse"]
        )
        self.assertTrue(isinstance(resp["entities"]["administrative"], list))
        self.assertTrue(len(resp["entities"]["administrative"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_entity_interface_1(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "entity-1.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        resp = await whoisit.entity_async("GOVI", rir="arin")
        self.assertEqual(resp["type"], "entity")
        self.assertEqual(resp["name"], "Govital Internet Inc.")
        self.assertEqual(resp["handle"], "GOVI")
        self.assertEqual(resp["parent_handle"], "")
        self.assertEqual(resp["rir"], "arin")
        self.assertEqual(resp["registration_date"], datetime(2001, 5, 8, 0, 0, tzinfo=tzoffset(None, -14400)))
        self.assertEqual(resp["expiration_date"], None)
        self.assertEqual(resp["url"], "https://rdap.arin.net/registry/entity/GOVI")
        self.assertEqual(resp["terms_of_service_url"], "https://www.arin.net/resources/registry/whois/tou/")
        self.assertEqual(resp["whois_server"], "whois.arin.net")
        self.assertEqual(
            resp["description"], ["http://www.govital.net\r", "Standard NOC hours are 10am to 6pm EST M-F"]
        )
        self.assertTrue(isinstance(resp["entities"]["technical"], list))
        self.assertTrue(len(resp["entities"]["technical"]) > 0)

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_save_bootstrap_data(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

        result = whoisit.save_bootstrap_data()
        self.assertTrue(isinstance(result, str))

        data = json.loads(result)
        self.assertTrue(bool(data))

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_save_bootstrap_data_no_json(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data(as_json=False)
        self.assertTrue(isinstance(data, dict))
        self.assertTrue(bool(data))

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_load_bootstrap_data(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data()
        self.assertTrue(isinstance(data, str))

        whoisit.clear_bootstrapping()

        whoisit.load_bootstrap_data(data)
        self.assertTrue(whoisit.is_bootstrapped())

    @pytest.mark.asyncio
    @pytest.mark.usefixtures("mock_httpx")
    async def test_load_bootstrap_data_no_json(self):
        load_sync_responses_to_httpx_mock(RESPONSES / "boostrap.yaml", self.httpx_mock)
        await whoisit.bootstrap_async()
        self.assertTrue(whoisit.is_bootstrapped())

        data = whoisit.save_bootstrap_data(as_json=False)
        self.assertTrue(isinstance(data, dict))

        whoisit.clear_bootstrapping()

        whoisit.load_bootstrap_data(data, from_json=False)
        self.assertTrue(whoisit.is_bootstrapped())



import json
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network
from time import time
from urllib.parse import urlsplit

from .errors import BootstrapError, UnsupportedError
from .logger import get_logger
from .overrides import iana_overrides
from .utils import (
    get_async_client,
    get_session,
    http_request,
    http_request_async,
    is_subnet_of,
)


log = get_logger('bootstrap')


class BaseBootstrap:

    # Where we get bootstrap data from
    BOOTSTRAP_URLS = {
        'asn': 'https://data.iana.org/rdap/asn.json',
        'dns': 'https://data.iana.org/rdap/dns.json',
        'ipv4': 'https://data.iana.org/rdap/ipv4.json',
        'ipv6': 'https://data.iana.org/rdap/ipv6.json',
        'object': 'https://data.iana.org/rdap/object-tags.json',
    }
    # RDAP by default must be over HTTPS, but can be toggled to allow HTTP
    SECURE_RDAP_PROTOCOLS = ('https',)
    INSECURE_RDAP_PROTOCOLS = ('http', 'https')
    # Servers to fall back to in the event of failing to resolve the correct server
    DEFAULT_RDAP_FALLBACKS = (
        'https://rdap.arin.net/registry/',
        'https://rdap.db.ripe.net/',
        'https://rdap.apnic.net/'
    )
    # Map of RIR RDAP endpoints
    RIR_RDAP_ENDPOINTS = {
        'afnic': 'https://rdap.nic.fr/',
        'afrinic': 'https://rdap.afrinic.net/rdap/',
        'arin': 'https://rdap.arin.net/registry/',
        'apnic': 'https://rdap.apnic.net/',
        'jpnic': 'https://jpnic.rdap.apnic.net/',
        'idnic': 'https://idnic.rdap.apnic.net/',
        'krnic': 'https://krnic.rdap.apnic.net/',
        'lacnic': 'https://rdap.lacnic.net/rdap/',
        'registro.br': 'https://rdap.registro.br/',
        'ripe': 'https://rdap.db.ripe.net/',
        'twnic': 'https://twnic.rdap.apnic.net/',
    }
    # Map of entity prefix and postfixes to RIRs, used to guess entity RIR
    RIR_ENTITY_PREFIXES = {
        'AFNIC': 'afnic',
        'AFRINIC': 'afrinic',
        'ARIN': 'arin',
        'AP': 'apnic',
        'JPNIC': 'jpnic',
        'KR': 'krnic',
        'ID': 'idnic',
        'LACNIC': 'lacnic',
        'BR': 'registro.br', # registro.br currently does not use entity prefixes
        'RIPE': 'ripe',
        'TW': 'twnic',
    }

    def __init__(self):
        self.bootstrap_parsers = {
            'asn': self.parse_asn_data,
            'dns': self.parse_dns_data,
            'ipv4': self.parse_ipv4_data,
            'ipv6': self.parse_ipv6_data,
            'object': self.parse_object_data,
        }
        self.rir_endpoints_by_domain = {}
        for name, url in self.RIR_RDAP_ENDPOINTS.items():
            url_parts = urlsplit(url)
            self.rir_endpoints_by_domain[url_parts.netloc] = name
        self._allow_insecure = False
        self._is_bootstrapped = False
        self._bootstrap_timestamp = 0
        self._expected_items = set(self.BOOTSTRAP_URLS.keys())
        self._data = {}
        self._parsed_data = {}
        self._use_iana_overrides = False
        self.clear_bootstrapping()

    def is_using_overrides(self):
        return self._use_iana_overrides

    def is_allowing_insecure_endpoints(self):
        return self._allow_insecure

    def is_bootstrapped(self):
        return self._is_bootstrapped

    def clear_bootstrapping(self):
        self._data = {}
        self._parsed_data = {}
        for k in self.BOOTSTRAP_URLS.keys():
            self._data[k] = {}
        self._is_bootstrapped = False
        self._use_iana_overrides = False
        log.debug('Cleared bootstrap data')

    def _bootstrap(self, overrides=False, allow_insecure=False):
        if self.is_bootstrapped():
            return True
        self._use_iana_overrides = bool(overrides)
        self._allow_insecure = bool(allow_insecure)
        items_loaded = set()
        for name, url in self.BOOTSTRAP_URLS.items():
            response = yield url
            yield
            if response.status_code != 200:
                raise BootstrapError(f'Failed to download bootstrap URL: {url}, got '
                                     f'non-200 response code: {response.status_code}')
            data = False
            try:
                data = response.json()
            except Exception as e:
                raise BootstrapError(f'Failed to parse data in URL: {url} '
                                     f'as JSON: {e}') from e
            if data:
                self._data[name] = data
                items_loaded.add(name)
        if items_loaded == self._expected_items:
            self._bootstrap_timestamp = int(time())
            self._is_bootstrapped = True
            self.parse_bootstrap_data()
            log.debug('Bootstrapped')
            return True
        else:
            items_missing = self._expected_items - items_loaded
            raise BootstrapError(f'Failed to load some bootstrap data, '
                                 f'missing data: {items_missing}')

    def save_bootstrap_data(self, as_json=True):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        rtn = {'timestamp': self._bootstrap_timestamp}
        for name, data in self._data.items():
            rtn[name] = data
        if as_json:
            return json.dumps(rtn)
        return rtn


    def load_bootstrap_data(self, data, overrides=False, allow_insecure=False, from_json=True):
        if self.is_bootstrapped():
            raise BootstrapError('Already bootstrapped, cannot load more data')
        if not isinstance(data, str) and from_json:
            raise BootstrapError(f'Unable to load bootstrap data, data must be a '
                                 f'string, got: {type(data)}')
        elif not isinstance(data, dict) and not from_json:
            raise BootstrapError(f'Unable to load bootstrap data, data must be a '
                                 f'dict, got: {type(data)}')
        self._use_iana_overrides = bool(overrides)
        self._allow_insecure = bool(allow_insecure)
        try:
            if from_json:
                data = json.loads(data)
        except Exception as e:
            raise BootstrapError(f'Unable to load bootstrap data, failed to parse '
                                 f'as JSON: {e}') from e
        timestamp = data.get('timestamp', None)
        if not isinstance(timestamp, int):
            raise BootstrapError(f'Unable to load bootstrap data, missing or '
                                 f'invalid timestamp')
        items_loaded = set()
        for name, item in data.items():
            if name == 'timestamp':
                continue
            self._data[name] = item
            items_loaded.add(name)
        if items_loaded == self._expected_items:
            self._bootstrap_timestamp = int(timestamp)
            self._is_bootstrapped = True
            self.parse_bootstrap_data()
            return True
        else:
            self.clear_bootstrapping()
            items_missing = self._expected_items - items_loaded
            raise BootstrapError(f'Failed to load some bootstrap data, '
                                 f'missing data: {items_missing}')

    def bootstrap_is_older_than(self, days):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        if not isinstance(days, int):
            raise BootstrapError(f'Days must be an integer, got: {type(days)}')
        day = 60 * 60 * 24
        age_in_seconds = int(time()) - self._bootstrap_timestamp
        age_in_days = age_in_seconds / day
        log.debug(f'Bootstrap data age in days: {age_in_days} > {days}')
        return age_in_days > days

    def parse_bootstrap_data(self):
        '''
            The bootstrap data itself is not organised in a way that can be quickly
            resolved. Parse each bootstrap objects data to make fast lookups possible.
        '''
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        self._parsed_data = {}
        for item, parser in self.bootstrap_parsers.items():
            log.debug(f'Parsing {item} data with parser {parser}')
            services = self._data[item].get('services', [])
            if not services:
                raise BootstrapError(f'Unable to parse "{item}" bootstrap data, '
                                     f'no "services" found')
            self._parsed_data[item] = parser(services)
        return True

    def validate_rdap_urls(self, urls):
        allowed_schemes = (self.INSECURE_RDAP_PROTOCOLS if self._allow_insecure
                           else self.SECURE_RDAP_PROTOCOLS)
        insecure_scheme = False
        validated_urls = []
        for url in urls:
            url = str(url).strip()
            url_parts = urlsplit(url)
            url_scheme = url_parts.scheme.lower()
            if url_scheme in allowed_schemes:
                validated_urls.append(url)
                insecure_scheme = url_scheme in self.INSECURE_RDAP_PROTOCOLS
        if not validated_urls:
            if insecure_scheme:
                insecure_str = (' (insecure scheme, try '
                                'whoisit.bootstrap(allow_insecure=True))')
            else:
                insecure_str = ''
            log.debug(f'No valid RDAP service URLs could be parsed '
                      f'from: {urls}{insecure_str}')
        return validated_urls

    def parse_asn_data(self, services):
        parsed = {}
        for selector, urls in services:
            validated_urls = self.validate_rdap_urls(urls)
            if not validated_urls:
                continue
            for asnrange in selector:
                parts = asnrange.split('-')
                num_parts = len(parts)
                if num_parts == 1:
                    range_start = int(parts[0])
                    range_end = int(parts[0])
                    parsed[(range_start, range_end)] = validated_urls
                elif num_parts == 2:
                    range_start = int(parts[0])
                    range_end = int(parts[1])
                    parsed[(range_start, range_end)] = validated_urls
                else:
                    raise BootstrapError(f'Invalid ASN range selector: {asnrange}')
        return parsed

    def parse_dns_data(self, services):
        parsed = {}
        for selector, urls in services:
            validated_urls = self.validate_rdap_urls(urls)
            if not validated_urls:
                continue
            for tld in selector:
                tld = tld.strip()
                parsed[tld] = validated_urls
        return parsed

    def parse_ipv4_data(self, services):
        parsed = {}
        for selector, urls in services:
            validated_urls = self.validate_rdap_urls(urls)
            if not validated_urls:
                continue
            for prefix in selector:
                network = IPv4Network(prefix, strict=True)
                parsed[network] = validated_urls
        return parsed

    def parse_ipv6_data(self, services):
        parsed = {}
        for selector, urls in services:
            validated_urls = self.validate_rdap_urls(urls)
            if not validated_urls:
                continue
            for prefix in selector:
                network = IPv6Network(prefix, strict=True)
                parsed[network] = validated_urls
        return parsed

    def parse_object_data(self, services):
        # Bootstrap entity information doesn't contain any mappings to RIR endpoints
        # so just ignore it and return an empty dict
        return {}

    def get_fallback_endpoints(self):
        return self.DEFAULT_RDAP_FALLBACKS

    def get_asn_endpoints(self, asn):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        if not isinstance(asn, int):
            raise BootstrapError('asn must be an int')
        for (range_start, range_end), endpoints in self._parsed_data['asn'].items():
            if range_start <= asn <= range_end:
                log.debug(f'Mapped ASN "{asn}" as between "{range_start}-{range_end}" '
                          f'and to endpoints: {endpoints}')
                return endpoints, True
        log.debug(f'Failed to map ASN "{asn}" to an endpoint, using a fallback')
        return self.get_fallback_endpoints(), False

    def get_dns_endpoints(self, tld):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        if not isinstance(tld, str):
            raise BootstrapError('tld must be an str')
        if self._use_iana_overrides:
            domain_overrides = iana_overrides.get('domain', {})
            override_endpoints = domain_overrides.get(tld, None)
            if override_endpoints:
                log.debug(f'Mapped TLD "{tld}" to override endpoints: '
                          f'{override_endpoints}')
                return override_endpoints, False
        endpoints = self._parsed_data['dns'].get(tld, None)
        if endpoints:
            log.debug(f'Mapped TLD "{tld}" to endpoints: {endpoints}')
            return endpoints, True
        # Domains have no fallback endpoints
        raise UnsupportedError(f'TLD "{tld}" has no known RDAP endpoint '
                               f'and is unsupported. It may be supported by '
                               f'overrides, try using whoisit.bootstrap(overrides=True)')

    def get_ipv4_endpoints(self, ipv4):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        if not isinstance(ipv4, (IPv4Address, IPv4Network)):
            raise BootstrapError('ipv4 must be a IPv4Address or a IPv4Network')
        if ipv4.is_private:
            raise BootstrapError(f'IPv4Address {ipv4} is private')
        if isinstance(ipv4, IPv4Address):
            for prefix, endpoints in self._parsed_data['ipv4'].items():
                if ipv4 in prefix:
                    log.debug(f'Mapped IPv4 address "{ipv4}" as in prefix "{prefix}" '
                              f'and to endpoints: {endpoints}')
                    return endpoints, True
        else:
            for prefix, endpoints in self._parsed_data['ipv4'].items():
                if ipv4 == prefix:
                    log.debug(f'Mapped IPv4 prefix "{ipv4}" exactly to endpoints: '
                              f'{endpoints}')
                    return endpoints, True
                elif is_subnet_of(ipv4, prefix):
                    log.debug(f'Mapped IPv4 prefix "{ipv4}" as a subnet of prefix '
                              f'"{prefix}" and to endpoints: {endpoints}')
                    return endpoints, True
        return self.get_fallback_endpoints(), False
    
    def get_ipv6_endpoints(self, ipv6):
        if not self.is_bootstrapped():
            raise BootstrapError('No bootstrap data is loaded')
        if not isinstance(ipv6, (IPv6Address, IPv6Network)):
            raise BootstrapError('ipv6 must be a IPv6Address or IPv6Network')
        if ipv6.is_private:
            raise BootstrapError(f'IPv6Address {ipv6} is private')
        if isinstance(ipv6, IPv6Address):
            for prefix, endpoints in self._parsed_data['ipv6'].items():
                if ipv6 in prefix:
                    log.debug(f'Mapped IPv6 address "{ipv6}" as in prefix "{prefix}" '
                              f'and to endpoints: {endpoints}')
                    return endpoints, True
        else:
            for prefix, endpoints in self._parsed_data['ipv6'].items():
                if ipv6 == prefix:
                    log.debug(f'Mapped IPv6 prefix "{ipv6}" exactly to endpoints: '
                              f'{endpoints}')
                    return endpoints, True
                elif is_subnet_of(ipv6, prefix):
                    log.debug(f'Mapped IPv6 prefix "{ipv6}" as a subnet of prefix '
                              f'"{prefix}" and to endpoints: {endpoints}')
                    return endpoints, True
        return self.get_fallback_endpoints(), False

    def get_entity_endpoints(self, entity):
        entity = entity.strip().upper()
        # Attempt to match a prefix or postfix to an RIR, for example many entities have
        # names such as RIPE-NAME or EXAMPLE-AP, we can use these prefixes and postfixes
        # to attempt to guess the correct RIR to query
        for part, rir_name in self.RIR_ENTITY_PREFIXES.items():
            if entity.startswith(f'{part}-') or entity.endswith(f'-{part}'):
                endpoint = self.RIR_RDAP_ENDPOINTS.get(rir_name)
                endpoints = [endpoint]
                log.debug(f'Mapped entity "{entity}" to RIR "{rir_name}" and to '
                          f'endpoints: {endpoints}')
                return endpoints, True
        # No match found, as querying a random RIR RDAP endpoint for a likely unknown
        # entity is almost certainly going to fail, raise it as unsupported
        raise UnsupportedError(f'Entity "{entity}" has no detectable RDAP endpoint, '
                               f'try specifying one manually with rir=...')

    def get_rir_endpoint(self, name):
        # allow 'ripencc' as an alias
        if name == 'ripencc':
            name = 'ripe'
        try:
            return self.RIR_RDAP_ENDPOINTS[name]
        except KeyError:
            raise BootstrapError(f'Invalid RIR endpoint name: {name}, must be '
                                 f'one of: {self.get_rir_endpoint_names()}')

    def get_rir_endpoint_names(self):
        return tuple(self.RIR_RDAP_ENDPOINTS.keys())

    def get_rir_name_by_endpoint_url(self, url):
        """
            A reverse lookup that maps endpoint URLs like https://rdap.arin.net/whatever
            to a name like 'arin'.
        """
        url_parts = urlsplit(url)
        try:
            return self.rir_endpoints_by_domain[url_parts.netloc]
        except KeyError:
            raise BootstrapError(f'Unknown endpoint URL: {url}')


class Bootstrap(BaseBootstrap):

    def __init__(self, session=None, allow_insecure_ssl=False, do_super_init=True):
        if do_super_init:
            BaseBootstrap.__init__(self)
        self.session = get_session(session, allow_insecure_ssl=allow_insecure_ssl)

    def bootstrap(self, overrides=False, allow_insecure=False):
        gen = self._bootstrap(overrides, allow_insecure)
        for url in gen:
            response = http_request(self.session, url)
            gen.send(response)


class BootstrapAsync(BaseBootstrap):

    def __init__(self, client=None, allow_insecure_ssl=False, do_super_init=True):
        if do_super_init:
            BaseBootstrap.__init__(self)
        self.client = get_async_client(client, allow_insecure_ssl)

    async def bootstrap_async(self, overrides=False, allow_insecure=False):
        gen = self._bootstrap(overrides, allow_insecure)
        for url in gen:
            response = await http_request_async(self.client, url)
            gen.send(response)


class _BootstrapMainModule(Bootstrap, BootstrapAsync):

    def __init__(self) -> None:
        Bootstrap.__init__(self, do_super_init=True)
        BootstrapAsync.__init__(self, do_super_init=False)



from urllib3.util import retry

try:
    from urllib3.util import create_urllib3_context
except ImportError:
    from urllib3.util.ssl_ import create_urllib3_context

import httpx
import requests
from urllib3.poolmanager import PoolManager

from .errors import QueryError, UnsupportedError
from .logger import get_logger
from .version import version


log = get_logger('utils')
user_agent = 'whoisit/{version}'
insecure_ssl_ciphers = 'ALL:@SECLEVEL=1'
http_timeout = 10                    # Maximum time in seconds to allow for an HTTP request
http_retry_statuses = [429]          # HTTP status codes to trigger a retry wih backoff
http_max_retries = 3                 # Maximum number of HTTP requests to retry before failing
http_pool_connections = 10           # Maximum number of HTTP pooled connections
http_pool_maxsize = 10               # Maximum HTTP pool connection size
async_http_max_connections = 100     # Maximum number of HTTP connections allowed for async client
async_max_keepalive_connections = 20 # Allow the connection pool to maintain keep-alive connections below this point
_default_session = {'secure': None, 'insecure': False}


def get_session_or_async_client(session_or_async_client=None, allow_insecure_ssl=False, is_async=False):
    """
        Creates and caches the default sessions, one for secure (default) SSL
        and one for SSL with an insecure cipher suite.
    """
    global _default_session
    if session_or_async_client:
        if allow_insecure_ssl:
            if not _default_session['insecure']:
                _default_session['insecure'] = session_or_async_client
            return _default_session['insecure']
        else:
            if not _default_session['secure']:
                _default_session['secure'] = session_or_async_client
            return _default_session['secure']
    else:
        if allow_insecure_ssl:
            if is_async:
                _default_session['insecure'] = create_async_client(allow_insecure_ssl=allow_insecure_ssl)
            else:
                _default_session['insecure'] = create_session(allow_insecure_ssl=allow_insecure_ssl)
            return _default_session['insecure']
        else:
            if is_async:
                _default_session['secure'] = create_async_client(allow_insecure_ssl=allow_insecure_ssl)
            else:
                _default_session['secure'] = create_session(allow_insecure_ssl=allow_insecure_ssl)
            return _default_session['secure']


def get_session(session=None, allow_insecure_ssl=False) -> requests.Session:
    return get_session_or_async_client(session, allow_insecure_ssl, is_async=False)


def get_async_client(client=None, allow_insecure_ssl=False) -> httpx.AsyncClient:
    return get_session_or_async_client(client, allow_insecure_ssl, is_async=True)


class InsecureSSLAdapter(requests.adapters.HTTPAdapter):
    """
        Custom adapter to permit insecure SSL connections.
    """

    def init_poolmanager(self, connections, maxsize, block=False):
        insecure_ssl_ciphersuite = create_urllib3_context(ciphers=insecure_ssl_ciphers)
        self.poolmanager = PoolManager(ssl_context=insecure_ssl_ciphersuite)


def create_session(allow_insecure_ssl=False):
    session = requests.session()
    session_retry = retry.Retry(total=http_max_retries,
                                status_forcelist=http_retry_statuses,
                                backoff_factor=1)
    whoisit_adapter = requests.adapters.HTTPAdapter(
        max_retries=session_retry,
        pool_connections=http_pool_connections,
        pool_maxsize=http_pool_maxsize)
    session.mount('https://', whoisit_adapter)
    if allow_insecure_ssl:
        session.mount('https://', InsecureSSLAdapter())
    return session


def create_async_client(allow_insecure_ssl=False):
    limits = httpx.Limits(max_connections=async_http_max_connections, max_keepalive_connections=async_max_keepalive_connections)
    retries = httpx.AsyncHTTPTransport(retries=http_max_retries, limits=limits)
    headers = {"User-Agent": user_agent.format(version=version)}
    verify = not allow_insecure_ssl
    client = httpx.AsyncClient(transport=retries, verify=verify, headers=headers, follow_redirects=True)
    return client


def http_request(session, url, method='GET', headers=None, data=None, *args, **kwargs):
    """
        Simple wrapper over requests. Allows for optionally downgrading SSL
        ciphers if required.
    """
    headers = headers or {}
    data = data or {}
    methods = ('GET',)
    if method not in methods:
        raise UnsupportedError(f'HTTP methods supported are: {methods}, got: {method}')
    headers['User-Agent'] = user_agent.format(version=version)
    log.debug(f'Making HTTP {method} request to {url}')
    try:
        if 'timeout' not in kwargs:
            kwargs['timeout'] = http_timeout
        return session.request(method, url, headers=headers, data=data, *args, **kwargs)
    except Exception as e:
        raise QueryError(f'Failed to make a {method} request to {url}: {e}') from e


async def http_request_async(client: httpx.AsyncClient, url, method='GET', headers=None, data=None, *args, **kwargs):
    """
        Simple wrapper over httpx.
    """
    headers = headers or {}
    data = data or {}
    methods = ('GET',)
    if method not in methods:
        raise UnsupportedError(f'HTTP methods supported are: {methods}, got: {method}')
    log.debug(f'Making async HTTP {method} request to {url}')
    try:
        if 'timeout' not in kwargs:
            kwargs['timeout'] = http_timeout
        return await client.request(method, url, headers=headers, data=data, *args, **kwargs)
    except Exception as e:
        raise QueryError(f'Failed to make a {method} request to {url}: {e}') from e


def is_subnet_of(network_a, network_b):
    a_len = network_a.prefixlen
    b_len = network_b.prefixlen
    return a_len >= b_len and network_a.supernet(a_len - b_len) == network_b


default_chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-'


def contains_only_chars(s, chars=default_chars):
    for c in s:
        if c not in chars:
            return False
    return True


def recursive_merge(d1, d2):
    '''
        Recursively merge two dictionaries. This is used to overlay subrequest
        data from related info RDAP endpoints over the top of primary RDAP
        request data. It has some special handling to account for related RDAP
        info having slightly different formats for events, notices, and remarks.
    '''
    for k, v in d2.items():
        if k in d1 and isinstance(d1[k], dict) and isinstance(v, dict):
            recursive_merge(d1[k], v)
        elif k == 'events' and isinstance(v, list):
            v1 = d1.get(k) or []
            d1[k] = recursive_merge_lists(v1, v, dedup_on='eventAction')
        elif k in {'notices', 'remarks'} and isinstance(v, list):
            v1 = d1.get(k) or []
            d1[k] = recursive_merge_lists(v1, v)
        elif v:
            d1[k] = v


def recursive_merge_lists(l1, l2, dedup_on='title'):
    list1 = {l[dedup_on]: l for l in l1 if dedup_on in l}
    list2 = {l[dedup_on]: l for l in l2 if dedup_on in l}
    recursive_merge(list1, list2)
    return list(list1.values())



import unittest
import json
from pathlib import Path
from ipaddress import IPv4Network, IPv4Address, IPv6Network, IPv6Address

import whoisit


BASE_DIR = Path(__file__).resolve().parent


class BootstrapTestCase(unittest.TestCase):

    maxDiff = None

    def setUp(self):
        whoisit.clear_bootstrapping()
        with open(BASE_DIR / 'data_bootstrap.json', 'rt') as f:
            self.bootstrap_data = f.read()

    def tearDown(self):
        whoisit.clear_bootstrapping()

    def test_bootstrap_saving(self):
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data)
        test_dict = json.loads(self.bootstrap_data)
        saved_data = whoisit.save_bootstrap_data()
        saved_dict = json.loads(saved_data)
        self.assertEqual(test_dict, saved_dict)
        whoisit.clear_bootstrapping()

    def test_bootstrap_loading(self):
        whoisit.clear_bootstrapping()
        with self.assertRaises(whoisit.errors.BootstrapError):
            whoisit.load_bootstrap_data('')
        with self.assertRaises(whoisit.errors.BootstrapError):
            whoisit.load_bootstrap_data('{"some":"dict"}')
        with self.assertRaises(whoisit.errors.BootstrapError):
            whoisit.load_bootstrap_data(12345)
        with self.assertRaises(whoisit.errors.BootstrapError):
            whoisit.load_bootstrap_data(b'test')
        whoisit.clear_bootstrapping()
        self.assertFalse(whoisit.is_bootstrapped())
        whoisit.load_bootstrap_data(self.bootstrap_data)
        self.assertTrue(whoisit.is_bootstrapped())
        test_dict = json.loads(self.bootstrap_data)
        for item in whoisit._bootstrap.BOOTSTRAP_URLS.keys():
            self.assertEqual(test_dict[item], whoisit._bootstrap._data[item])
        whoisit.clear_bootstrapping()

    def test_bootstrap_asn_parser(self):
        test_rdap_url = 'https://rdap.example.com/'
        test_data = [[['1-2', '6-7', '9'], [test_rdap_url]]]
        expected = {
            (1, 2): [test_rdap_url],
            (6, 7): [test_rdap_url],
            (9, 9): [test_rdap_url],
        }
        self.assertEqual(whoisit._bootstrap.parse_asn_data(test_data), expected)

    def test_bootstrap_dns_parser(self):
        test_rdap_url = 'https://rdap.example.com/'
        test_data = [[['a', 'b', 'c'], [test_rdap_url]]]
        expected = {
            'a': [test_rdap_url],
            'b': [test_rdap_url],
            'c': [test_rdap_url],
        }
        self.assertEqual(whoisit._bootstrap.parse_dns_data(test_data), expected)

    def test_bootstrap_ipv4_parser(self):
        test_rdap_url = 'https://rdap.example.com/'
        test_data = [[['10.0.0.0/8', '127.0.0.0/8'], [test_rdap_url]]]
        expected = {
            IPv4Network('10.0.0.0/8'): [test_rdap_url],
            IPv4Network('127.0.0.0/8'): [test_rdap_url],
        }
        self.assertEqual(whoisit._bootstrap.parse_ipv4_data(test_data), expected)

    def test_bootstrap_ipv6_parser(self):
        test_rdap_url = 'https://rdap.example.com/'
        test_data = [[['2001:1400::/22', '2001:200::/23'], [test_rdap_url]]]
        expected = {
            IPv6Network('2001:1400::/22'): [test_rdap_url],
            IPv6Network('2001:200::/23'): [test_rdap_url],
        }
        self.assertEqual(whoisit._bootstrap.parse_ipv6_data(test_data), expected)

    def test_bootstrap_ipv6_parser(self):
        test_rdap_url = 'https://rdap.example.com/'
        test_data = [[['does', 'nothing', 'yet'], [test_rdap_url]]]
        expected = {}
        self.assertEqual(whoisit._bootstrap.parse_object_data(test_data), expected)

    def test_bootstrap_full_parsing(self):

        # Load bootstrap data
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data)

        # ASN endpoint tests
        test = 327691
        expected = ['https://rdap.afrinic.net/rdap/'], True
        self.assertEqual(whoisit._bootstrap.get_asn_endpoints(test), expected)
        test = 4567
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_asn_endpoints(test), expected)
        test = 206236
        expected = ['https://rdap.db.ripe.net/'], True
        self.assertEqual(whoisit._bootstrap.get_asn_endpoints(test), expected)
        test = 2047
        expected = ['https://rdap.db.ripe.net/'], True
        self.assertEqual(whoisit._bootstrap.get_asn_endpoints(test), expected)
        # One of the random fallbacks for an invalid ASN
        test = 59438593485903
        random_servers, exact_match = whoisit._bootstrap.get_asn_endpoints(test)
        self.assertTrue(isinstance(random_servers[0], str))
        self.assertEqual(exact_match, False)
        # One of the random fallbacks for an invalid ASN
        test = -1
        random_servers, exact_match = whoisit._bootstrap.get_asn_endpoints(test)
        self.assertTrue(isinstance(random_servers[0], str))
        self.assertEqual(exact_match, False)

        # Domain endpoint tests
        test = 'com'
        expected = ['https://rdap.verisign.com/com/v1/'], True
        self.assertEqual(whoisit._bootstrap.get_dns_endpoints(test), expected)
        test = 'no'
        expected = ['https://rdap.norid.no/'], True
        self.assertEqual(whoisit._bootstrap.get_dns_endpoints(test), expected)
        test = 'dev'
        expected = ['https://www.registry.google/rdap/'], True
        self.assertEqual(whoisit._bootstrap.get_dns_endpoints(test), expected)
        # Invalid TLDs should raise an error
        test = 'testtesttesttesttesttesttesttesttesttesttest'
        with self.assertRaises(whoisit.errors.UnsupportedError):
            whoisit._bootstrap.get_dns_endpoints(test)

        # IPv4 prefix endpoint tests
        test = IPv4Address('1.1.1.1')
        expected = ['https://rdap.apnic.net/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv4_endpoints(test), expected)
        test = IPv4Network('1.1.1.0/24')
        expected = ['https://rdap.apnic.net/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv4_endpoints(test), expected)
        test = IPv4Address('8.8.8.8')
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv4_endpoints(test), expected)
        test = IPv4Network('8.8.8.0/24')
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv4_endpoints(test), expected)
        with self.assertRaises(whoisit.errors.BootstrapError):
            # Private IP addresses can't have a public RDAP endpoint at all
            whoisit._bootstrap.get_ipv4_endpoints(IPv4Address('127.0.0.1'))

        # IPv6 prefix endpoint tests
        test = IPv6Address('2606:4700:4700::1111')
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv6_endpoints(test), expected)
        test = IPv6Network('2606:4700::/32')
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv6_endpoints(test), expected)
        test = IPv6Address('2001:67c:2e8:22::c100:68b')
        expected = ['https://rdap.db.ripe.net/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv6_endpoints(test), expected)
        test = IPv6Network('2001:67c:2e8::/48')
        expected = ['https://rdap.db.ripe.net/'], True
        self.assertEqual(whoisit._bootstrap.get_ipv6_endpoints(test), expected)
        with self.assertRaises(whoisit.errors.BootstrapError):
            # Private IP addresses can't have a public RDAP endpoint at all
            whoisit._bootstrap.get_ipv6_endpoints(IPv6Address('fc00::1'))

        # Entity endpoint tests with prefixes
        test = 'ENTITY-RIPE'
        expected = ['https://rdap.db.ripe.net/'], True
        self.assertEqual(whoisit._bootstrap.get_entity_endpoints(test), expected)
        test = 'ENTITY-AP'
        expected = ['https://rdap.apnic.net/'], True
        self.assertEqual(whoisit._bootstrap.get_entity_endpoints(test), expected)
        test = 'ARIN-ENTITY'
        expected = ['https://rdap.arin.net/registry/'], True
        self.assertEqual(whoisit._bootstrap.get_entity_endpoints(test), expected)
        # Entities without a prefix or postfix we can parse are unsupported
        with self.assertRaises(whoisit.errors.UnsupportedError):
            whoisit._bootstrap.get_entity_endpoints('ANYTHING')

        # Clean up
        whoisit.clear_bootstrapping()

    def test_override_endpoint(self):

        # Load bootstrap data
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data)

        # Invalid RIR endpoint name
        with self.assertRaises(whoisit.errors.BootstrapError):
            whoisit._bootstrap.get_rir_endpoint('test')

        # Check the RIRs are all supported
        for name, endpoint in whoisit._bootstrap.RIR_RDAP_ENDPOINTS.items():
            self.assertEqual(whoisit._bootstrap.get_rir_endpoint(name), endpoint)

        # Check the RIR names are valid
        expected = ('afnic', 'afrinic', 'arin', 'apnic', 'jpnic', 'idnic', 'krnic',
                    'lacnic', 'registro.br', 'ripe', 'twnic')
        self.assertEqual(whoisit._bootstrap.get_rir_endpoint_names(), expected)

        # Clean up
        whoisit.clear_bootstrapping()

    def test_iana_overrides(self):

        # Test iana_overrides are a dict
        from whoisit.overrides import iana_overrides
        self.assertIsInstance(iana_overrides, dict)

        # Test bootstrapping with overrides disabled
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data)
        self.assertFalse(whoisit._bootstrap.is_using_overrides())
        with self.assertRaises(whoisit.errors.UnsupportedError):
            # .de has no RDAP entry in IANA data currently, this should error
            whoisit._bootstrap.get_dns_endpoints('de')

        # Test bootstrapping with overrides enabled
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data, overrides=True)
        self.assertTrue(whoisit._bootstrap.is_using_overrides())
        # .de has endpoint overrides
        override_endpoints, match = whoisit._bootstrap.get_dns_endpoints('de')
        self.assertEqual(override_endpoints[0], 'https://rdap.denic.de/')
        self.assertFalse(match)
        # Test for SLD overrides
        override_endpoints, match = whoisit._bootstrap.get_dns_endpoints('gl')
        self.assertEqual(override_endpoints[0], 'https://rdap.centralnic.com/gl/')
        self.assertFalse(match)
        override_endpoints, match = whoisit._bootstrap.get_dns_endpoints('co.gl')
        self.assertEqual(override_endpoints[0], 'https://rdap.centralnic.com/co.gl/')
        self.assertFalse(match)

    def test_insecure_scheme_endpoints(self):

        # Test secure HTTPS endpoints, the default
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data, allow_insecure=False)
        self.assertFalse(whoisit._bootstrap.is_allowing_insecure_endpoints())
        with self.assertRaises(whoisit.errors.UnsupportedError):
            # .uz only has an insecure HTTP RDAP entry in IANA data
            whoisit._bootstrap.get_dns_endpoints('uz')

        # Test bootstrapping with insecure endpoints allowed
        whoisit.clear_bootstrapping()
        whoisit.load_bootstrap_data(self.bootstrap_data, allow_insecure=True)
        self.assertTrue(whoisit._bootstrap.is_allowing_insecure_endpoints())
        # .uz should now have a valid RDAP endpoint over HTTP
        override_endpoints, match = whoisit._bootstrap.get_dns_endpoints('uz')
        self.assertEqual(override_endpoints[0], 'http://cctld.uz:9000/')
        self.assertTrue(match)
